﻿#include "emulator/Emulator.hpp"

#include <conio.h>
#include <windows.h>

#include <iomanip>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <unordered_set>
#include <locale.h>
using namespace std;

class Application {
    static SHORT round(double v) //
    {
        return static_cast<SHORT>(std::round(v));
    }

    void print(Location location, char c)
    {
        // Напечатать символ на координатах
        const auto x = round(location.x + 1);
        const auto y = round(location.y + 1);

        if (x > 0 && y > 0) {
            COORD pos = { x, y };
            SetConsoleCursorPosition(mOutput, pos);
            std::cout << c;
        }
    }

    int getchar()
    {
        // Читаем один символ если такие там есть
        if (_kbhit())
            return _getch();
        return 0;
    }

public:
    explicit Application(Mode mode)
        : mMode(mode)
        , mOutput(GetStdHandle(STD_OUTPUT_HANDLE)){}
    explicit Application(Mode mode, double sensitivity)
        : mMode(mode)
        , mOutput(GetStdHandle(STD_OUTPUT_HANDLE))
    {
        if (sensitivity <= 0 || sensitivity > 1)
            sensitivity = 0.5;
        else
            this->sensitivity = sensitivity;
    }
    
    //пукнт 3
    bool run(State& state)
    {
        // ручное управление
        if (mMode == Mode::HandsOn) {
            draw(state);

            
            auto& car = state.getCar();
            //при этих условиях машина будет сохранять скорость и направление движения
            car.turnSteeringWheel(0);
            car.setGear(Gear::Neutral);
            car.pressBrakePedal(0);

            //считываем символ
            char c = getchar();
            if (c == 'w' || c == 'W') {
                //если необходимо развернуться устанавливаем соответствующее значение
                if (direction != 1)
                    car.setGear(Gear::Forward);
                direction = 1;

            } else if (c == 's' || c == 'S') {
                if (direction != -1)
                    car.setGear(Gear::Reverse);
                direction = -1;
            } else if (c == 'e' || c == 'E') {
                //нельзя ускоряться на нейтрали, поэтому устанавливаем сохраненное направление
                if (direction == 1)
                    car.setGear(Gear::Forward);
                else if (direction == -1)
                    car.setGear(Gear::Reverse);
                car.pressGasPedal(sensitivity);
            } else if (c == 'q' || c == 'Q') {
                car.pressBrakePedal(sensitivity);
            } else if (c == 'a' || c == 'A')
                car.turnSteeringWheel(-sensitivity);
            else if (c == 'd' || c == 'D')
                car.turnSteeringWheel(sensitivity);

            return !state.isCollision() && !state.isDestination();
        } 
        else {
            draw(state);
            oneLineStraight(state);
        }
    }

    bool atRoad(State& state, int index, Location location) { 
        auto lane = state.getRoad().getLanes()[index];
        const auto direction = lane.getDirection();
        const auto startPoint = lane.getLocation();
        const Location endPoint = { startPoint.x + direction.dx, startPoint.y + direction.dy };

        const auto size = 0;

        if (startPoint.x <= location.x - size && location.x + size <= endPoint.x && //
            startPoint.y <= location.y - size && location.y + size <= endPoint.y) {
            return true;
        }
    }

    //2.1
    bool oneLineStraight(State& state) { 
        auto& car = state.getCar(); 
        if (state.getCurrentTime() == 0) {
            if (car.getLocation().x > car.getDestination().center.x) {
                car.setGear(Gear::Reverse);
                car.pressGasPedal(0.05);
            } else {
                car.setGear(Gear::Forward);
                //так как accelerate <= 1, a minSpeed >= 0.01 таким образом получим гарантированное минимальную скорость
                car.pressGasPedal(0.01);
            }
        } else if (car.getLocation().isInRange(car.getDestination().center, car.getDestination().radius)) {
            car.setGear(Gear::Neutral); 
            car.pressBrakePedal(1);
        }
        return !state.isCollision() && !state.isDestination();
    }

    //2.1
    bool oneLineWithoutSigns(State& state) { 
        auto& car = state.getCar();
        car.setGear(Gear::Forward);

        switch (state.getCurrentTime()) {
        case 0: { //ищем дорогу на которой стоим
            int lanes = state.getRoad().getLanes().size();
            for (int i = 0; i < lanes; i++) {
                if (atRoad(state, i, car.getLocation())) {
                    curLane = i;
                    break;
                }
            }

            //проверяем, как далеко можно газануть в направлении, в котором стоит машина
            auto& lane = state.getRoad().getLanes()[curLane];
            auto endOfLane = lane.getLocation().moved(lane.getDirection(), 1);
            koeffForAcceleration = 1;
            while (!atRoad(state, curLane,
                car.getLocation().moved(car.getDirection(), car.getMaxSpeed(Gear::Forward) * koeffForAcceleration)))
                koeffForAcceleration /= 2;
            //газуем
            car.setGear(Gear::Forward);
            car.pressGasPedal(koeffForAcceleration);
            break;
        }
        case 1: {

            //смотрим, на сколько смогли разогнаться за секунду
            forwardAccleration = car.getCurrentSpeed() / koeffForAcceleration;
            startSpeed = car.getCurrentSpeed();
            car.setGear(Gear::Neutral);
            car.pressBrakePedal(1);
            break;
        }
        case 2: { 
            //смотрим, на сколько получилось затормозить
            deceleration = startSpeed - car.getCurrentSpeed();
            break;
        }
        default:
            if (car.getDestination().center.x - car.getLocation().x)
            break;
        }
        return !state.isDestination() && !state.isCollision();
    }

    //пункт 1
    void draw(State& state)
    {
        // Очищаем старые символы
        {
            for (auto location : mLocations)
                print(location, ' ');

            mLocations.clear();
            notDrawAt.clear();
        }

        auto road = state.getRoad();

        //выводим скорость
        double speed = state.getCar().getCurrentSpeed();
        string speed_str = to_string(speed);
        for (int i = 1; i < 10; i++) {
            COORD pos = { i, 1 };
            SetConsoleCursorPosition(mOutput, pos);
            cout << ' '; 
        }
        for (int i = 1; i <= speed_str.length(); i++) {
            COORD pos = { i, 1 };
            SetConsoleCursorPosition(mOutput, pos);
            cout << speed_str[i - 1]; 
        }

        //рисуем знаки
        vector<Location> signsToDraw;
        auto trafficSigns = road.getSigns();
        for (auto sign : trafficSigns) {
            auto offset = sign.getLocation();
            int index = sign.getLaneIndex();
            auto curLine = road.getLanes()[index];
            Location location
                = { curLine.getLocation().x + offset, curLine.getLocation().y + curLine.getDirection().dy / 2 };

             SetConsoleTextAttribute(mOutput, 12);
            if (sign.getType() == TrafficSign::Type::MaximumSpeedLimit) {
                print(location, 'T'); // top
            }
            else
                print(location, 'B'); // bottom
            SetConsoleTextAttribute(mOutput, 15);
            notDrawAt[round(location.x)].insert(round(location.y));
        }

        //рисуем чужие машины
        auto cars = state.getVisibleVehicles();
        vector<Location> carsToDraw;
        SetConsoleTextAttribute(mOutput, 13);
        for (auto car : cars) {
            auto location = car.getLocation();
            print(location, 'o');
            mLocations.push_back(location);
            notDrawAt[round(location.x)].insert(round(location.y));
        }
        SetConsoleTextAttribute(mOutput, 15);

        //точка назначения
        SetConsoleTextAttribute(mOutput, 10);
        print(state.getCar().getDestination().center, 'F');
        SetConsoleTextAttribute(mOutput, 15);
        notDrawAt[round(state.getCar().getDestination().center.x)].insert(round(state.getCar().getDestination().center.y));

        //рисуем свою машину
        auto myCarLocation = state.getCar().getLocation();
        SetConsoleTextAttribute(mOutput, 14);
        print(myCarLocation, '@');
        SetConsoleTextAttribute(mOutput, 15);
        mLocations.push_back(myCarLocation);
        notDrawAt[round(myCarLocation.x)].insert(round(myCarLocation.y));

        //рисуем дорогу, не рисуем ее поверх знаков
        vector<Lane> lanes = road.getLanes();
        for (auto lane : lanes) {
            auto begin = lane.getLocation();
            int width = lane.getDirection().dy;
            int length = lane.getDirection().dx;
            for (int i = begin.x; i <= begin.x + length; i++) {
                Location location = { i, begin.y };
                if (notDrawAt.find(round(location.x)) == notDrawAt.end()
                    || notDrawAt[round(location.x)].find(round(location.y)) == notDrawAt[round(location.x)].end()) 
                    print(location, '-');
                location.y = begin.y + width; 
               if (notDrawAt.find(round(location.x)) == notDrawAt.end()
                    || notDrawAt[round(location.x)].find(round(location.y)) == notDrawAt[round(location.x)].end()) 
                    print(location, '-');
            }
            for (int j = begin.y + 1; j < begin.y + width; j++) {
                Location location = {begin.x, j}; 
               if (notDrawAt.find(round(location.x)) == notDrawAt.end()
                    || notDrawAt[round(location.x)].find(round(location.y)) == notDrawAt[round(location.x)].end()) 
                    print(location, '|'); 
                location.x = begin.x + length; 
                if (notDrawAt.find(round(location.x)) == notDrawAt.end()
                    || notDrawAt[round(location.x)].find(round(location.y)) == notDrawAt[round(location.x)].end()) 
                    print(location, '|');
            }
        }

        
    }
   
 private:
    HANDLE mOutput;
    Mode mMode;
    std::vector<Location> mLocations;
    double forwardAccleration; 
    double startSpeed;
    double deceleration; 
    double reverseAccleration; 
    double rotationRate;
    Direction start_direction;
    unordered_map<SHORT, unordered_set<SHORT>> notDrawAt;
    int direction = 0; //текущее направление движения: 1 - вперед, -1 - назад
    double sensitivity; //чувствительность управления
    int curLane;
    double koeffForAcceleration; 
    };

int main(int argc, char** argv)
{
    try {
        Emulator emulator(argc, argv);
        //второй параметр - чувствительность ручного управления - параметр для pressGasPedal и т.д
        Application app(emulator.getMode(), 0.8);
        emulator.run([&app](State& state) { return app.run(state); });
        SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), COORD { 1, 30 });
        return 0;
    } catch (const std::runtime_error& error) {
        std::cerr << error.what() << std::endl;
        return 1;
    }
}


